# reference: https://discuss.python.org/t/how-to-make-a-pure-python-wheel-for-2-7-3-6/71858/2
import sysconfig
import os

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


class CustomHook(BuildHookInterface):
    def initialize(self, version, build_data):
        if self.target_name not in ('wheel', 'bdist'):
            return
        if os.environ.get("WHEEL_PLATFORM") is not None:
            platform_tag = os.environ["WHEEL_PLATFORM"]
        else:
            # https://peps.python.org/pep-0425/#platform-tag
            platform_tag = sysconfig.get_platform().replace('-','_').replace('.','_')

        build_data['tag'] = f'py3-none-{platform_tag}'



from typing import Optional
from pathlib import Path
from shenzi.discover.plat import get_platform


def search_lib(name: str) -> Optional[Path]:
    from shenzi.discover.search.linux import search_linux
    from shenzi.discover.search.mac import search_mac

    plat = get_platform()
    if plat == "linux":
        return search_linux(name)
    elif plat == "mac":
        result = search_mac(name)
        return Path(result) if result else None
    else:
        # unknown
        raise Exception(f"platform {plat} is unknown")


def cffi_find_library(name: str) -> Optional[Path]:
    return search_lib(name)


def ctypes_cdll_find_library(name: str) -> Optional[Path]:
    return search_lib(name)



import subprocess
import os
from typing import Optional
from pathlib import Path


def ldconfig_find(name: str) -> Optional[Path]:
    try:
        output = subprocess.check_output([
            '/sbin/ldconfig', '-p'
        ], env={**os.environ, 'LANG': 'C', 'LC_ALL': 'C'})
        output = output.decode('utf-8')
    except Exception:
        return None
    candidates = _find_in_output(name, output)
    if candidates:
        return candidates
    return None

def _find_in_output(name: str, output: str) -> Optional[Path]:
    candidates: list[str] = []
    for line in output.splitlines():
        candidate = _get_candidate(name, line)
        if candidate:
            candidates.append(candidate)
    # Prefer exact match
    for candidate in candidates:
        if candidate.endswith(name):
            p = Path(candidate)
            if p.exists():
                return p
    # Fallback: first existing candidate
    for candidate in candidates:
        p = Path(candidate)
        if p.exists():
            return p
    return None

def _get_candidate(name: str, line: str) -> Optional[str]:
    comps = line.split('=>')
    if len(comps) < 2:
        return None
    candidate = comps[1].strip()
    if name in candidate:
        return candidate
    return None




from functools import partial
import json
import os
import atexit
from typing import Optional
from shenzi.discover.all_libs import get_libs
from shenzi.discover.callbacks import cffi_dlopen_callback, ctypes_cdll_callback
from shenzi.discover.imports import add_import_callback, register_import_watcher
from shenzi.discover.python_props import get_python_props
from shenzi.discover.monkeypatch import kwarg_else_arg, try_monkey_patch
from shenzi.discover.types import *

DUMP_LOC_ENV_VAR = "SHENZI_JSON"
DEFAULT_LOC = "shenzi.json"


def monkey_patch_dlopen(loads):
    try_monkey_patch(
        "ctypes",
        ["cdll", "LoadLibrary"],
        partial(ctypes_cdll_callback, loads=loads),
        kwarg_else_arg("name", 0),
    )
    try_monkey_patch(
        "ctypes",
        ["CDLL", "__init__"],
        partial(ctypes_cdll_callback, loads=loads),
        kwarg_else_arg("name", 1),
    )
    try_monkey_patch(
        "cffi",
        ["api", "FFI", "dlopen"],
        partial(cffi_dlopen_callback, loads=loads),
        kwarg_else_arg("name", 1),
    )
    try_monkey_patch(
        "cffi",
        ["FFI", "dlopen"],
        partial(cffi_dlopen_callback, loads=loads),
        kwarg_else_arg("name", 1),
    )


def _validate_prepared_loads(loads: dict[LocalLoad, LoadParams]):
    for local_load in loads:
        if not os.path.isabs(local_load.path):
            raise Exception(
                f"shenzi critical failure, found a path while searching libraries which is not absolute, path={local_load.path}"
            )


def main_exit_handler(pkgs_to_skip: list[str], loads):
    import sys
    from pathlib import Path

    site_pkgs = [Path(p) for p in sys.path]
    prefixes_to_skip = [p / skip for p in site_pkgs for skip in pkgs_to_skip]
    prefixes_to_skip = [p for p in prefixes_to_skip if p.exists()]
    prefixes_to_skip = [str(p) for p in prefixes_to_skip]

    exit_handler(prefixes_to_skip, loads)


def exit_handler(prefixes_to_skip: list[str], loads):
    _validate_prepared_loads(loads)

    dump_loc = os.environ.get(DUMP_LOC_ENV_VAR, DEFAULT_LOC)
    payload = ShenziDiscovery(
        loads=[
            Load(path=load.path, symlinks=list(param.symlinks), kind=load.kind)
            for load, param in loads.items()
        ],
        libs=[Lib(path=lib) for lib in get_libs()],
        python=get_python_props(),
        skip=Skip(prefixes=prefixes_to_skip, libs=[]),
        env={str(k): str(v) for k, v in os.environ.items()},
        bins=[],
    )
    with open(dump_loc, "w") as f:
        json.dump(payload.to_dict(), f)


_ENABLED_DISCOVERY = False


def shenzi_init_discovery(skip: Optional[list[str]] = None):
    from multiprocessing import Manager
    _manager = Manager()
    LOADS = _manager.dict()  # type: ignore

    if not skip:
        skip = []
    global _ENABLED_DISCOVERY
    if _ENABLED_DISCOVERY:
        return
    _ENABLED_DISCOVERY = True
    monkey_patch_dlopen(LOADS)
    register_import_watcher(partial(add_import_callback, loads=LOADS))
    atexit.register(lambda: main_exit_handler(skip, LOADS))



from shenzi.discover.plat import get_platform


def get_libs() -> list[str]:
    plat = get_platform()
    if plat == "mac":
        return _get_mac_libs()
    elif plat == "linux":
        return _get_linux_libs()
    else:
        return []


def _get_mac_libs() -> list[str]:
    from shenzi.discover.all_libs.macho import get_dyld_finder

    dyld_finder = get_dyld_finder()
    all_dylibs = dyld_finder()
    return all_dylibs


def _get_linux_libs():
    libs = set()
    with open("/proc/self/maps", "r") as f:
        for line in f:
            parts = line.strip().split()
            if len(parts) >= 6 and parts[-1].endswith(".so"):
                libs.add(parts[-1])
    return sorted(libs)


import os
from pathlib import Path
from typing import Callable, Optional
from shenzi.discover.search import cffi_find_library, ctypes_cdll_find_library
from shenzi.discover.types import LoadParams, LocalLoad


def cffi_dlopen_callback(name: str, loads) -> None:
    _std_load_callback(name, loads, _cffi_dlopen, "cffi.FFI.dlopen", True)


def ctypes_cdll_callback(name: str, loads) -> None:
    _std_load_callback(name, loads, _ctypes_cdll_dlopen, "ctypes.CDLL", True)


def _std_load_callback(
    name: str,
    loads: dict[LocalLoad, LoadParams],
    params_getter: Callable[
        [str, dict[LocalLoad, LoadParams]], Optional[tuple[LocalLoad, LoadParams]]
    ],
    patched_id: str,
    strict: bool,
) -> None:
    maybe_result = params_getter(name, loads)
    if not maybe_result:
        if not strict:
            return
        else:
            raise Exception(
                f"shenzi: failed in finding dynamic library, name={name}, dlopen_called_from={patched_id}"
            )
    local_load, params = maybe_result
    loads[local_load] = params


def _cffi_dlopen(
    name: str, loads: dict[LocalLoad, LoadParams]
) -> Optional[tuple[LocalLoad, LoadParams]]:
    return _std_dlopen(name, loads, cffi_find_library)


def _ctypes_cdll_dlopen(
    name: str, loads: dict[LocalLoad, LoadParams]
) -> Optional[tuple[LocalLoad, LoadParams]]:
    return _std_dlopen(name, loads, ctypes_cdll_find_library)


def _std_dlopen(
    name: str,
    loads: dict[LocalLoad, LoadParams],
    fd_lib: Callable[[str], Optional[Path]],
) -> Optional[tuple[LocalLoad, LoadParams]]:
    lib = fd_lib(name)
    if not lib:
        print("could not find lib for", name)
        return None
    if not lib.exists():
        raise Exception(
            f"got a path from find_library which does not exist, search_term={name} path={lib} finder={fd_lib.__qualname__}"
        )

    lib = os.path.realpath(lib)
    symlinks_to_add = get_symlinks(name, lib)

    local_load = LocalLoad(path=lib, kind="dlopen")
    if local_load in loads:
        loaded_params = loads[local_load]
        symlinks_to_add = symlinks_to_add | loaded_params.symlinks

    return local_load, LoadParams(symlinks=symlinks_to_add)


def get_symlinks(name: str, full_path: str) -> set[str]:
    _name = Path(name)
    _full_path = Path(full_path)
    if _name.name == _full_path.name:
        return set()
    else:
        return set([_name.name])



def get_dyld_finder():
    import ctypes
    import ctypes.util

    _libdyld = ctypes.CDLL(ctypes.util.find_library("libdyld.dylib"))

    _libdyld._dyld_image_count.restype = ctypes.c_uint
    _libdyld._dyld_get_image_name.restype = ctypes.c_char_p

    def wrapped():
        count = _libdyld._dyld_image_count()
        res = []
        for i in range(count):
            image_name = _libdyld._dyld_get_image_name(i)
            res.append(image_name.decode())
        return res

    return wrapped


def get_search_terms_for_dylib(name: str) -> list[str]:
    return _get_search_terms_using_by_ctypes(name) + _adhoc_search_term_heuristics(name)


def _adhoc_search_term_heuristics(name: str) -> list[str]:
    res = _get_search_terms_with_suffixes(name)
    if not name.startswith("lib"):
        res.extend(_get_search_terms_with_suffixes(f"lib{name}"))
    return res


def _get_search_terms_with_suffixes(name: str) -> list[str]:
    # TODO: find all the dyld conventions for suffixes (which files are same kinda suffixes)
    res = [
        f"{name}.dylib" if not name.endswith(".dylib") else None,
        f"{name}.0.dylib" if not name.endswith(".0.dylib") else None,
        f"{name}-0.dylib" if not name.endswith("-0.dylib") else None,
        f"{name}.so" if not name.endswith(".so") else None,
        f"{name}.so.0" if not name.endswith(".so.0") else None,
    ]
    return [term for term in res if term is not None]


def _get_search_terms_using_by_ctypes(name: str) -> list[str]:
    import os

    return [os.path.basename(p) for p in get_ctypes_search_paths(name)]


def get_ctypes_search_paths(name: str) -> list[str]:
    # copied from ctypes.macholib.dyld
    from itertools import chain
    import os
    import sys
    from ctypes.macholib.dyld import (
        dyld_image_suffix_search,
        dyld_override_search,
        dyld_executable_path_search,
        dyld_default_search,
    )

    possible = [
        "@executable_path/../lib/lib%s.dylib" % name,
        "lib%s.dylib" % name,
        "%s.dylib" % name,
        "%s.framework/%s" % (name, name),
    ]

    res = []
    for name in possible:
        iterator = dyld_image_suffix_search(
            chain(
                dyld_override_search(name, os.environ),
                dyld_executable_path_search(name, sys.executable),
                dyld_default_search(name, os.environ),
            ),
            os.environ,
        )
        res.extend(iterator)

    return res



from dataclasses import dataclass
from typing import Any, Literal, Optional

LoadKind = Literal["extension", "dlopen"]

@dataclass(frozen=True)
class LocalLoad:
    kind: LoadKind
    path: str


@dataclass
class LoadParams:
    symlinks: set[str]


@dataclass(frozen=True)
class Load:
    path: str
    kind: LoadKind
    symlinks: list[str]

    def to_dict(self) -> dict[str, Any]:
        return {"path": self.path, "symlinks": self.symlinks, "kind": self.kind}

@dataclass(frozen=True)
class Lib:
    path: str
    
    def to_dict(self) -> dict[str, Any]:
        return {"path": self.path}

@dataclass(frozen=True)
class Bin:
    path: str
    
    def to_dict(self) -> dict[str, Any]:
        return {"path": self.path}


@dataclass(frozen=True)
class Version:
    major: int
    minor: int
    abi_thread: str
    
    def to_dict(self) -> dict[str, Any]:
        return {
            "major": self.major,
            "minor": self.minor,
            "abi_thread": self.abi_thread
        }


@dataclass(frozen=True)
class Sys:
    prefix: str
    exec_prefix: str
    platlibdir: str
    version: Version
    path: list[str]
    executable: str
    
    def to_dict(self) -> dict[str, Any]:
        return {
            "prefix": self.prefix,
            "exec_prefix": self.exec_prefix,
            "platlibdir": self.platlibdir,
            "version": self.version.to_dict(),
            "path": self.path,
            "executable": self.executable
        }


@dataclass(frozen=True)
class Python:
    sys: Sys
    main: str
    allowed_packages: Optional[list[str]]
    cwd: str
    
    def to_dict(self) -> dict[str, Any]:
        return {"sys": self.sys.to_dict(), "main": self.main, "allowed_packages": self.allowed_packages, "cwd": self.cwd}


@dataclass(frozen=True)
class Skip:
    prefixes: list[str]
    libs: list[str]
    
    def to_dict(self) -> dict[str, Any]:
        return {"prefixes": self.prefixes, "libs": self.libs}



@dataclass(frozen=True)
class ShenziDiscovery:
    loads: list[Load]
    libs: list[Lib]
    bins: list[Bin]
    python: Python
    skip: Skip
    env: dict[str, str]
    
    def to_dict(self) -> dict[str, Any]:
        return {
            "loads": [load.to_dict() for load in self.loads],
            "libs": [lib.to_dict() for lib in self.libs],
            "bins": [bin.to_dict() for bin in self.bins],
            "python": self.python.to_dict(),
            "skip": self.skip.to_dict(),
            "env": self.env,
        }


from pathlib import Path
import sys
import importlib.abc
from shenzi.discover.types import LoadParams, LocalLoad

def register_import_watcher(add_ext_callback):
    class ImportWatcher(importlib.abc.MetaPathFinder):
        def find_spec(self, fullname, path, target=None):
            # We delegate finding this import to other finders
            # if any finder returns a spec, pass a `shenzi.types.Module` to callback
            spec = None
            for finder in sys.meta_path:
                if finder is not self:
                    try:
                        spec = finder.find_spec(fullname, path, target)
                        if spec is not None:
                            break
                    except (ImportError, AttributeError):
                        continue

            if spec is not None and spec.origin:
                if self._is_dyn_lib(spec.origin):
                    add_ext_callback(spec.origin)
            return None

        def _is_dyn_lib(self, path):
            return path.endswith(".dylib") or path.endswith(".so")

    sys.meta_path.insert(0, ImportWatcher())


def add_import_callback(path: str, loads: dict[LocalLoad, LoadParams]) -> None:
    p = Path(path)
    if not p.exists():
        raise Exception(f"import watcher found a path which does not exist: {path}")
    p = p.resolve()
    local_load = LocalLoad(kind="extension", path=str(p))
    params = LoadParams(symlinks=set([]))
    loads[local_load] = params



import ctypes.util
from typing import Optional
from pathlib import Path


def search_mac(name: str) -> Optional[str]:
    if Path(name).exists():
        return name
    return ctypes.util.find_library(name)




import importlib
from typing import Any
import logging


logger = logging.getLogger(__name__)

def _get_element(mod, attrs):
    el = mod
    for attr in attrs:
        el = getattr(el, attr)
    return el


def _set_element(mod, attrs, el):
    parent = _get_element(mod, attrs[:-1])
    setattr(parent, attrs[-1], el)


def _monkey_patch(mod, attrs, add_lib_callback, args_to_path):
    try:
        original_fn = _get_element(mod, attrs)

        def new_fn(*args, **kwargs):
            path = args_to_path(*args, **kwargs)
            result = original_fn(*args, **kwargs)
            if not result or not path:
                return result
            logger.info("adding %s", path)
            add_lib_callback(path)
            return result

        _set_element(mod, attrs, new_fn)
    except AttributeError as ex:
        print("failed in patching", mod, attrs, ex)


def try_monkey_patch(mod, attrs, add_lib_callback, args_to_path):
    try:
        mod_ = importlib.import_module(mod)
        _monkey_patch(mod_, attrs, add_lib_callback, args_to_path)
        print("patched", mod, attrs)
    except ImportError as ex:
        print("failed in importing", mod, ex)


def kwarg_else_arg(var: str, i: int) -> Any:
    """return a function which tries to find `kwargs[var]` or `args[i]` from its arguments

    When we monkey patch a function like `dlopen`, we want to infer the path that `dlopen` was trying to open
    We intercept its args and given *args and **kwargs, try to find the potential path the args point to

    if `var` is in kwargs, that is returned, else `args[i]` is returned

    NOTE: this returns a relative path if the args themselves get a relative path, see what we can do about it
    Maybe also try to infer from return types?
    """
    def _f(*a, **kw):
        return kw[var] if var in kw else a[i]

    return _f


import platform
from pathlib import Path
import shutil
import subprocess
from typing import Optional


def get_rpaths(path: Path) -> tuple[list[str], list[str]]:
    # uses objdump
    # TODO: use the shenzi cli that we ship with the library only
    # guaranteed success
    objdump = shutil.which("objdump")
    if not objdump:
        return ([], [])

    try:
        output = subprocess.check_output(["objdump", "-p", str(path)])
        output = output.decode("utf-8")
    except Exception:
        return ([], [])

    # output looks like this
    # Dynamic Section:
    # NEEDED               libpthread.so.0
    # RPATH                $ORIGIN/../lib

    rpath, runpath = "", ""
    for line in output.splitlines():
        line = line.strip()
        if line.startswith("RPATH"):
            parts = line.split(None)
            if len(parts) > 1:
                rpath = parts[1]
        elif line.startswith("RUNPATH"):
            parts = line.split(None, 1)
            if len(parts) > 1:
                runpath = parts[2]

    return (rpath.split(":") if rpath else [], runpath.split(":") if runpath else [])


def parse_rpath(rpath: str, object_path: Path) -> Optional[Path]:
    parent_path_str = str(object_path.parent)
    # TODO: use getauxval instead of this
    is_64_bit = platform.architecture()[0] == "64bit"
    lib = "lib64" if is_64_bit else "lib"
    at_platform = platform.machine()

    rpath_expanded = (
        rpath.replace(r"$ORIGIN", parent_path_str)
        .replace(r"${ORIGIN}", parent_path_str)
        .replace(r"$LIB", lib)
        .replace(r"${LIB}", lib)
        .replace(r"$PLATFORM", at_platform)
        .replace(r"${PLATFORM}", at_platform)
    )

    path = Path(rpath_expanded)
    if path.exists():
        return path.resolve()
    return None





def main() -> None:
    import importlib.resources as resources
    import sys
    import os
    binary_name = "shenzi" if sys.platform != "win32" else "shenzi.exe"
    with resources.path("shenzi.bin", "shenzi") as shenzi:
        if sys.platform == "win32":
            import subprocess
            completed_process = subprocess.run([shenzi, *sys.argv[1:]])
            sys.exit(completed_process.returncode)
        else:
            os.execvp(shenzi, [shenzi, *sys.argv[1:]])



import sys
import os
from shenzi.discover.types import Python, Sys, Version
from pathlib import Path


def get_python_props() -> Python:
    realpath = os.path.realpath
    
    abi_thread = "t" if hasattr(sys, "abiflags") and "t" in sys.abiflags else ""
    python_path = [realpath(p) for p in sys.path if Path(p).exists()]
    return Python(
        sys=Sys(
            prefix=realpath(sys.base_prefix),
            exec_prefix=realpath(sys.base_exec_prefix),
            platlibdir=sys.platlibdir,
            version=Version(
                major=sys.version_info.major,
                minor=sys.version_info.minor,
                abi_thread=abi_thread,
            ),
            path=python_path,
            executable=realpath(sys.executable),
        ),
        main=str(Path(sys.argv[0]).resolve()),
        allowed_packages=None,
        cwd=str(Path.cwd()),
    )



from ctypes.util import find_library
from pathlib import Path
import sys
from typing import Optional
import os

from shenzi.discover.search.linux.ldconfig import ldconfig_find
from shenzi.discover.search.linux.rpath import get_rpaths, parse_rpath


SYS_RPATHS = {}


def search_linux(name: str) -> Optional[Path]:
    ld_library_path, ld_preload = _get_library_paths()
    cwd = Path.cwd()
    rpaths, runpaths = _get_sys_rpaths()
    ctypes_find = find_library(name)
    name = ctypes_find if ctypes_find else name

    return _search(
        name, rpaths, runpaths, ld_preload, ld_library_path, cwd
    )

def _get_sys_rpaths():
    exe = Path(sys.executable)
    rpaths, runpaths = get_rpaths(exe)

    rpaths = [parse_rpath(r, exe) for r in rpaths]
    rpaths = [r for r in rpaths if r]

    runpaths = [parse_rpath(r, exe) for r in runpaths]
    runpaths = [r for r in runpaths if r]

    return rpaths, runpaths

def _get_library_paths():
    ld_library_path = os.environ.get("LD_LIBRARY_PATH", "")
    ld_preload = os.environ.get("LD_PRELOAD", "")

    ld_library_path_dirs = [
        Path(p) for p in ld_library_path.split(":") if p and Path(p).exists()
    ]
    ld_preload_libs = [
        Path(p) for p in ld_preload.split(":") if p and Path(p).exists()
    ]

    return ld_library_path_dirs, ld_preload_libs

def _search(
    name: str,
    dt_rpaths: list[Path],
    dt_runpaths: list[Path],
    ld_preload: list[Path],
    ld_library_path: list[Path],
    cwd: Path,
) -> Optional[Path]:
    # 1. search as a path
    path = _search_name_as_path(name, cwd)
    if path:
        return path

    # 2. search LD_PRELOAD
    found = _find_in_dirs(name, ld_preload)
    if found:
        return found

    # 3. search DT_RPATH and extra_rpaths if DT_RUNPATH is empty
    if not dt_runpaths:
        found = _find_in_dirs(name, dt_rpaths)
        if found:
            return found

    # 4. search LD_LIBRARY_PATH
    found = _find_in_dirs(name, ld_library_path)
    if found:
        return found

    # 5. search DT_RUNPATH
    found = _find_in_dirs(name, dt_runpaths)
    if found:
        return found

    # 6. fallback, ask ldconfig
    found = ldconfig_find(name)
    if found:
        return found

    # 7. fallback to standard library dirs
    std_lib_dirs = [
        Path("/lib64"),
        Path("/lib"),
        Path("/usr/lib64"),
        Path("/usr/lib"),
    ]
    found = _find_in_dirs(name, std_lib_dirs)
    if found:
        return found

    return None

def _find_in_dirs(file_name: str, dirs: list[Path]) -> Optional[Path]:
    for dir in dirs:
        candidate = dir / file_name
        if candidate.exists():
            return candidate
    return None


def _search_name_as_path(name: str, cwd: Path) -> Optional[Path]:
    if "/" not in name:
        return None
    p = Path(name)
    if p.is_absolute():
        return p if p.exists() else None
    else:
        candidate = cwd / p
        return candidate if candidate.exists() else None


import sys
from typing import Literal


Platform = Literal["linux", "darwin", "unknown"]

def get_platform():
    if sys.platform.startswith("linux"):
        return "linux"
    elif sys.platform == "darwin":
        return "mac"
    else:
        return "unknown"